/**
 * GMyth Library
 *
 * @file gmyth/gmyth_query.c
 * 
 * @brief <p> GMythQuery class provides a wrapper for accessing
 * the libmysqlclient funtions.
 *
 * Copyright (C) 2006 INdT - Instituto Nokia de Tecnologia.
 * @author Leonardo Sobral Cunha <leonardo.cunha@indt.org.br>
 *
 *//*
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "gmyth_query.h"
#include "gmyth_settings.h"
#include "gmyth_context.h"

static void gmyth_query_class_init          (GMythQueryClass *klass);
static void gmyth_query_init                (GMythQuery *object);

static void gmyth_query_dispose  (GObject *object);
static void gmyth_query_finalize (GObject *object);

static void gmyth_query_print_error (MYSQL *conn, char *message);

G_DEFINE_TYPE(GMythQuery, gmyth_query, G_TYPE_OBJECT)
    
static void
gmyth_query_class_init (GMythQueryClass *klass)
{
	GObjectClass *gobject_class;

    gobject_class = (GObjectClass *) klass;
	
    gobject_class->dispose  = gmyth_query_dispose;
    gobject_class->finalize = gmyth_query_finalize;	
}

static void
gmyth_query_init (GMythQuery *gmyth_query)
{
	GMythSettings *gmyth_settings = gmyth_context_get_settings ();
	
    gmyth_query->opt_host_name = gmyth_settings_get_backend_hostname(gmyth_settings);
    gmyth_query->opt_user_name = gmyth_settings_get_username(gmyth_settings);
    gmyth_query->opt_password = gmyth_settings_get_password(gmyth_settings);
    gmyth_query->opt_db_name = gmyth_settings_get_dbname(gmyth_settings);

    /* initialize connection handler */
    gmyth_query->conn = mysql_init (NULL);
    
    if (!(gmyth_query->conn))
    	g_warning ("[%s] MSQL structure not initialized", __FUNCTION__);

}

static void
gmyth_query_dispose  (GObject *object)
{
    GMythQuery *gmyth_query = GMYTH_QUERY (object);
    
    /* disconnect from server */
	gmyth_query_disconnect (gmyth_query);
	
	G_OBJECT_CLASS (gmyth_query_parent_class)->dispose (object);
}

static void
gmyth_query_finalize (GObject *object)
{
	g_signal_handlers_destroy (object);

    G_OBJECT_CLASS (gmyth_query_parent_class)->finalize (object);
}

/** Creates a new instance of GMythQuery.
 * 
 * @return a new instance of GMythQuery.
 */
GMythQuery*
gmyth_query_new ()
{
    GMythQuery *sql_query = GMYTH_QUERY (g_object_new(GMYTH_QUERY_TYPE, NULL));

    return sql_query;
}

/** Connects to the Mysql database in the backend. The backend address
 * is loaded from the GMythSettings instance.
 * 
 * @param gmyth_query the GMythEPG instance to be connected.
 * @return true if connection was success, false if failed.
 */
gboolean
gmyth_query_connect (GMythQuery *gmyth_query) 
{
    char *opt_host_name;
    char *opt_user_name;
    char *opt_password;
    char *opt_db_name;

    assert(gmyth_query);
    
    opt_host_name   = (gmyth_query->opt_host_name ? gmyth_query->opt_host_name->str : NULL);
    opt_db_name     = (gmyth_query->opt_db_name ? gmyth_query->opt_db_name->str : NULL);
    opt_user_name   = (gmyth_query->opt_user_name ? gmyth_query->opt_user_name->str : NULL);
    opt_password    = (gmyth_query->opt_password ? gmyth_query->opt_password->str : NULL);

    
    if (gmyth_query->conn == NULL) {
        gmyth_query_print_error (NULL, "mysql_init() failed (probably out of memory)");
        return FALSE;
    }
    
    /* connect to server */
    if (mysql_real_connect (gmyth_query->conn, opt_host_name, 
            opt_user_name, opt_password, opt_db_name, 
            0, NULL, 0) == NULL) {
            	
        gmyth_query_print_error (gmyth_query->conn, "mysql_real_connect() failed");
		return FALSE;
    }

    g_debug ("[%s] Connection to Mysql server succeeded!", __FUNCTION__);
    
    return TRUE;
}

/** Disconnects from the Mysql database in the backend.
 * 
 * @param gmyth_query the GMythQuery instance to be disconnected
 * @return true if disconnection was success, false if failed.
 */
gboolean
gmyth_query_disconnect (GMythQuery *gmyth_query) 
{
    assert(gmyth_query);
    
    /* TODO: Check how to return error */
    g_debug ("[%s] Closing gmyth_query->conn", __FUNCTION__);
    mysql_close (gmyth_query->conn);

	return TRUE;	
}

static void
gmyth_query_print_error (MYSQL *conn, char *message)
{
    fprintf (stderr, "%s\n", message);
    
    if (conn != NULL) {
#if MYSQL_VERSION_ID >= 40101
        fprintf (stderr, "Error %u (%s): %s\n",
                mysql_errno (conn), mysql_sqlstate(conn), mysql_error (conn));
#else
        fprintf (stderr, "Error %u: %s\n",
               mysql_errno (conn), mysql_error (conn));
#endif
    }
}

/** Sends the given query to the backend returning the query result as
 * MYSQL_RES pointer.
 * 
 * FIXME: this function is returning NULL whether any error happens
 * or no rows are returned (e.g. UPDATE or REPLACE).
 * 
 * @param gmyth_query the GMythQuery instance.
 * @param stmt_str the query text.
 * @return the MYSQL_RES result pointer or NULL if any error happens.
 */
MYSQL_RES*
gmyth_query_process_statement (GMythQuery *gmyth_query, char *stmt_str)
{
    MYSQL_RES *res_set;
    
    assert(gmyth_query);
    
    g_debug ("[%s] Running mysql query %s", __FUNCTION__, stmt_str);

    if (gmyth_query == NULL)
    	return NULL;
    
    /* the statement failed */
    if (mysql_query (gmyth_query->conn, stmt_str) != 0) {
        gmyth_query_print_error (gmyth_query->conn, "Could not execute statement");
        return NULL;
    }

    /* the statement succeeded; determine whether it returned data */
    res_set = mysql_store_result (gmyth_query->conn);
    if (res_set) {
        return res_set;
    } else if (mysql_field_count (gmyth_query->conn) == 0) {
             g_debug ("%lu rows affected\n",
                     (unsigned long) mysql_affected_rows (gmyth_query->conn));
    } else {
            gmyth_query_print_error (gmyth_query->conn, "Could not retrieve result set");
    }
    
    return NULL;
}
